/*
 * Copyright 2000, International Business Machines Corporation and others.
 * All Rights Reserved.
 * 
 * This software has been released under the terms of the IBM Public
 * License.  For details, see the LICENSE file in the top-level source
 * directory or online at http://www.openafs.org/dl/license10.html
 */

#include <winver.h>

1 VERSIONINFO
BEGIN
    BLOCK "StringFileInfo"
    /* FileVersion must be specified or data won't appear on properties tab */
    BEGIN
        BLOCK "040904E4"
        BEGIN
            VALUE "CompanyName",        "OpenAFS Project", "\0"
            VALUE "LegalCopyright",     "Copyright \251 IBM Corporation and others, 1998, 1999, 2000, 2001.", "\0"
            VALUE "ProductName",        "OpenAFS for Windows", "\0"
            VALUE "ProductVersion",     "1.1", "\0"
            VALUE "FileVersion",        "1.1", "\0"
            VALUE "FileDescription",    AFS_VERINFO_FILE_DESCRIPTION, "\0"
            VALUE "Build",              AFS_VERINFO_BUILD, "\0"
        END
    END
END
