# (C) Copyright Transarc Corporation 1989
# Licensed Materials - Property of Transarc
# All Rights Reserved
#
# Package prototype for a staff machine using the kernel dynamic loader.
#

#
# Base components
#
%include ${wsadmin}/lib/devel
%include ${wsadmin}/src/admin
%include ${wsadmin}/lib/base.generic

#
# Machine-specific components
#
%ifdef	rt_r4
%include ${wsadmin}/lib/rt_r4.readonly
%include ${wsadmin}/lib/rt_r4.generic.dkload
%include ${wsadmin}/lib/rt_r4.generic.dev
#%else
#%ifdef	pmax_31
#%include ${wsadmin}/lib/pmax_31.readwrite
#%include ${wsadmin}/lib/pmax_31.dkload.generic
#%include ${wsadmin}/lib/pmax_31.generic.dev
#%else
#%ifdef	aix_22
#%include ${wsadmin}/lib/tree.readonly
#%include ${wsadmin}/lib/aix.generic
#%include ${wsadmin}/lib/aixdevice.generic
#%else
#%include ${wsadmin}/lib/tree.readonly
#%include ${wsadmin}/lib/tree.generic
#%include ${wsadmin}/lib/device.generic
#%endif	aix_22
#%endif  pmax_31
%endif	rt_r4
