/*
 * Copyright 1988 by the Massachusetts Institute of Technology.
 *
 * For copying and distribution information, please see the file
 * <mit-cpyright.h>.
 *
 * Declaration of debugging flag.
 */

#include <afsconfig.h>
#include <afs/param.h>

RCSID("$Header: /cvs/openafs/src/des/debug_decl.c,v 1.3 2001/07/12 19:58:34 shadow Exp $");

#include <mit-cpyright.h>
int	des_debug	= 0;
