#ifndef _PARAM_ALPHA_LINUX2216_22_USR_H_
#define _PARAM_ALPHA_LINUX2216_22_USR_H_

/* In user space the AFS_LINUX20_ENV should be sufficient. In the kernel,
 * it's a judgment call. If something is obviously sparc64 specific, use that
 * #define instead. Note that "20" refers to the linux 2.0 kernel. The "2"
 * in the sysname is the current version of the client. This takes into
 * account the perferred OS user space configuration as well as the kernel.
 */

#define UKERNEL                        1       /* user space kernel */
#define AFS_ENV                        1
#define AFS_USR_LINUX20_ENV    1
#define AFS_USR_LINUX22_ENV    1
#define AFS_NONFSTRANS 1

#define AFS_MOUNT_AFS "afs"    /* The name of the filesystem type. */
#define AFS_SYSCALL 338
#define AFS_64BIT_IOPS_ENV  1
#define AFS_64BIT_ENV       1
#define AFS_NAMEI_ENV     1   /* User space interface to file system */
#include <afs/afs_sysnames.h>

#define AFS_USERSPACE_IP_ADDR 1
#define RXK_LISTENER_ENV 1
#define AFS_GCPAGS             0       /* if nonzero, garbage collect PAGs */


/* Machine / Operating system information */
#define SYS_NAME       "alpha_linux_22"
#define SYS_NAME_ID    SYS_NAME_ID_alpha_linux_22
#define AFSLITTLE_ENDIAN    1
#define AFS_HAVE_FFS        1       /* Use system's ffs. */
#define AFS_HAVE_STATVFS    0  /* System doesn't support statvfs */
#define AFS_VM_RDWR_ENV            1   /* read/write implemented via VM */

#define        afsio_iov       uio_iov
#define        afsio_iovcnt    uio_iovcnt
#define        afsio_offset    uio_offset
#define        afsio_seg       uio_segflg
#define        afsio_fmode     uio_fmode
#define        afsio_resid     uio_resid
#define        AFS_UIOSYS      1
#define        AFS_UIOUSER     UIO_USERSPACE
#define        AFS_CLBYTES     MCLBYTES
#define        AFS_MINCHANGE   2
#define        VATTR_NULL      usr_vattr_null

#define AFS_DIRENT
#ifndef CMSERVERPREF
#define CMSERVERPREF
#endif

#endif /* _PARAM_ALPHA_LINUX2216_22_USR_H_ */
